part of dart_mason;

class Component{
  String name;
  Transform _transform;

  materials.MaterialInstance _material;
  Set<Mesh> _meshes = new Set<Mesh>();
  List<Component> _components = new List<Component>();
  List<Component> get children => _components;

  bool _hasChanged = true;
  bool get hasChanged => _hasChanged;

  /// Matrix set by model import process.
  Matrix4 _importTransform = new Matrix4.identity();

  factory Component({materials.MaterialInstance material}) {
    var c = new Component._("", [], []);
    c._material = material;
    return c;
  }

  factory Component._inner([String name = ""]) {
    var component = new Component._(name, [], []);
    return component;
  }

  Component._(this.name, List<Mesh> meshes, List<Component> children) {
    transform = new Transform();
    for (var m in meshes) {
      addMesh(m);
    }
    for (var child in children) {
      addComponent(child);
    }
  }

  addMesh(Mesh mesh) {
    _meshes.add(mesh);
  }

  /// Adds a child [Component]
  addComponent(Component component) {
    if (component != null) {
      _components.add(component);
      component.transform.parent = transform;
    } else {
      throw "Component cannot be null.";
    }
  }

  /// Adds [List<Component>] as children
  addAll(List<Component> components) {
    if (components != null) {
      for (var c in components) {
        addComponent(c);
      }
    } else {
      throw "Component cannot be null.";
    }
  }

  Transform get transform => _transform;

  set transform(Transform transform) {
    _transform = transform;
  }

  _computeTransform(Matrix4 parent) {
    return transform.matrix * parent * _importTransform;
  }

  /// Components that have been imported via a 3d model format such as FBX have
  /// a transform related to that scene. This method resets that to be
  /// at the origin, no scale, and no rotation.
  zeroPosition() {
    _importTransform.setFrom(new Matrix4.identity());
  }

  Material get material => _material;
  set material(Material value) {
    _material = value;
  }

  setOffsetTransform(Float32List list) {
    _importTransform.copyFromArray(list);
  }

  String toString() {
    var output = "Component[name:${name}]";
    return output;
  }

  /// Prints returns tree of child [Component]
  /// @alpha
  String printTree() {
    return _printTreeLevel(0);
  }

  String _printTreeLevel([int level = 0]) {
    var currentLevel = level;
    level++;
    var output = " " * (currentLevel) + "Component[name:${name}]";
    output += "\n";
    for (var c in children) {
      output += " " * (currentLevel + 1) + c._printTreeLevel(level);
    }
    return output;
  }

  /// Finds first child [Component] by component [name].
  Component findChild(String name) {
    for (var c in children) {
      if (c.name == name) {
        return c;
      }
    }
    for (var c in children) {
      return c.findChild(name);
    }
    return null;
  }

  /// Finds all child [Component]'s with [name].
  List<Component> findChildren(String name) {
    List<Component> found = [];
    for (var c in children) {
      if (c.name == name) {
        found.add(c);
        found.addAll(c.findChildren(name));
      }
    }
    return found;
  }

  /// Mark's the [Component] as changed. This will trigger [Renderable]'s
  /// to be recalculated on [Scene].build()
  void markedChanged() {
    _hasChanged = true;
  }

  /// Keeps a temporary list of Renderables.
  List<Renderable> _cachedList;
  Matrix4 _cachedTransform;
  List<Renderable> _build(Matrix4 transform, [bool recalc = false]) {
    if (_hasChanged || recalc) {
      _cachedTransform = _computeTransform(transform);
  }
    var list = new List<Renderable>();
    for (var mesh in _meshes) {
      var uniforms = new List<UniformValue>();
      if(material != null){
        uniforms.addAll(material.build());
        uniforms
          .add(new UniformValue("model", _cachedTransform, UniformType.MAT4));
        list.add(new Renderable(material.shader, mesh, uniforms));
      }
    }
    for (var c in _components) {
      list.addAll(c._build(_cachedTransform, _hasChanged));
    }
    _hasChanged = false;
    _cachedList = list;
    return _cachedList;
  }
}
