part of dart_mason;

class UniformMap {
  Map<String, dynamic> _values = {};

  UniformMap();

  setVector3(String key, Vector3 value) {
    _values[key] = value;
  }

  setFloat(String key, double value) {
    _values[key] = value;
  }

  addArray(String key, List array) {
    _values[key] = array;
  }

  setTexture(String key, Texture texture) {
    _values[key] = texture;
  }

  addMap(String key, UniformMap map) {
    _values[key] = map;
  }
}

class Material extends UniformMap {
  static Material DEFAULT = new Material(Shader.DEFAULT);
  final Shader shader;

  Map<String, UniformMap> _keyedMaps = new Map<String, UniformMap>();
  Set<UniformMap> _maps = new Set<UniformMap>();

  /// Creates a new Material with given [Shader]
  /// You can add referenced [UniformMap]s or non referenced [UniformMap]s
  Material(this.shader,
      {Map<String, UniformMap> referencedMaps, List<UniformMap> maps}) {
    if(maps != null){
      _maps.addAll(maps);
    }
    if(referencedMaps != null){
      _keyedMaps.addAll(referencedMaps);
    }
  }

  /// Add's a [UniformMap] as a dependency
  /// Manually set uniforms will override any
  /// uniforms inherited from added maps.
  /// An optional [reference] can be set for later reference
  addUniformMap(UniformMap map, {String reference}) {
    if (reference != null) {
      _keyedMaps[reference] = map;
    } else {
      _maps.add(map);
    }
  }

  /// Create's a copy of a material. Dependent [UniformMap]s will be shared.
  /// Material specific uniforms will be copied.
  Material clone() {
    var material =
        new Material(shader, referencedMaps: _keyedMaps, maps: _maps.toList());
    _values.forEach((key, value) {
      material._values[key] = value;
    });
    return material;
  }

  /// Remove's a [UniformMap] as a dependency.
  removeUniformMap(String reference) {
    _keyedMaps.remove(reference);
  }

  /// Remove's all uniform maps. Useful in scenarios where [UniformMap]
  /// was added without a [reference].
  _clearUniformMaps() {}

  List<UniformValue> build() {
    var build = new Map<String, UniformValue>();
    //We first want to create uniform values for all of the deps
    for (var m in _maps) {
      m._values.forEach((key, value) {
        String uniform_key = key;
        if (value is UniformMap) {
          _buildMap(build, uniform_key + ".", value);
        } else if (value is List) {
          _buildList(build, uniform_key, value);
        } else {
          _buildValue(build, "", key, value);
        }
      });
    }
    //We then override any values that were added untop.
    _values.forEach((key, value) {
      String uniform_key = key;
      if (value is UniformMap) {
        _buildMap(build, uniform_key + ".", value);
      } else if (value is List) {
        _buildList(build, uniform_key, value);
      } else {
        _buildValue(build, "", key, value);
      }
    });

    return new List.from(build.values);
  }

  _buildMap(Map<String, UniformValue> build, String parentKey, UniformMap map) {
    map._values.forEach((key, value) {
      String uniform_key = "${parentKey}.${key}";
      if (value is UniformMap) {
        _buildMap(build, uniform_key, value);
      } else if (value is List) {
        _buildList(build, uniform_key, value);
      } else {
        _buildValue(build, parentKey, key, value);
      }
    });
  }

  _buildValue(Map<String, UniformValue> build, String parentKey, key, value) {
    var finalKey = "${parentKey}${key}";
    if (value is Vector3) {
      build[finalKey] = new UniformValue(finalKey, value, UniformType.VEC3);
    }
    if (value is double) {
      build[finalKey] = new UniformValue(finalKey, value, UniformType.FLOAT);
    }
    if (value is Texture) {
      build[finalKey] = new UniformValue(finalKey, value, UniformType.TEXTURE);
    }
  }

  _buildList(Map<String, UniformValue> build, String parentKey, List list) {
    var map = new Map<String, UniformValue>();
    if (list[0] is UniformMap) {
      for (var i = 0; i < list.length; i++) {
        _buildMap(build, "${parentKey}[$i].", list[i]);
      }
    } else {
      // It's just plain values
      for (var i = 0; i < list.length; i++) {
        var key = "${parentKey}[${i}]";
        _buildValue(build, key, "", list[i]);
//        map[key] = new UniformValue(key, list[i], UniformType.VEC3);
      }
    }
  }
}

class Shader {
  final int _id;
  static int _next_id = 0;
  static Shader DEFAULT = new Shader._(1);
  static Shader PBR = new Shader._(2);

  factory Shader() {
    return new Shader._(++_next_id);
  }

  Shader._(this._id);
}

/// Helper function used in c++ code to be able to set the texture.
void setTextureByUrl(Material material, String name, String url) {
  Texture.FromUrl(url).then((texture) {
    material.setTexture("material.${name}", texture);
  });
}
