part of dart_mason;

class _Vertex {
  Vector3 position;
  Vector3 normal;
  Vector3 tangent;
  Vector3 bitangent;
  Vector2 uv;

  static int _size = 14;
  static int CalcBufferSize(int numVertices) {
    return _size * numVertices;
  }

  _Vertex(ByteBuffer buffer, int vertexOffset) {
    int bufferOffset = vertexOffset * (_size * Float32List.BYTES_PER_ELEMENT);
    position = new Vector3.fromBuffer(buffer, bufferOffset);
    bufferOffset += (3 * Float32List.BYTES_PER_ELEMENT);
    normal = new Vector3.fromBuffer(buffer, bufferOffset);
    bufferOffset += (3 * Float32List.BYTES_PER_ELEMENT);
    tangent = new Vector3.fromBuffer(buffer, bufferOffset);
    bufferOffset += (3 * Float32List.BYTES_PER_ELEMENT);
    bitangent = new Vector3.fromBuffer(buffer, bufferOffset);
    bufferOffset += (3 * Float32List.BYTES_PER_ELEMENT);
    uv = new Vector2.fromBuffer(buffer, bufferOffset);
  }
}

enum MeshAttribute {
  POSITION,
  NORMAL,
  TANGENT,
  TEXCOORD_0,
  TEXCOORD_1,
  COLOR_0,
//  JOINTS_0,
//  JOINTS_1,
//  JOINTS_2,
//  JOINTS_3,
//  WEIGHTS_0,
//  WEIGHTS_1,
//  WEIGHTS_2,
//  WEIGHTS_3,
}

enum AccessorType { VEC2, VEC3, VEC4 }

enum ComponentType {
  BYTE,
  UNSIGNED_BYTE,
  SHORT,
  UNSIGNED_SHORT,
  UNSIGNED_INT,
  FLOAT,
}

class AttributeDescription {
  final MeshAttribute name;
  final AccessorType accessor;
  final ComponentType component;

  static AttributeDescription POSITION = new AttributeDescription(
      MeshAttribute.POSITION, AccessorType.VEC3, ComponentType.FLOAT);
  static AttributeDescription NORMAL = new AttributeDescription(
      MeshAttribute.NORMAL, AccessorType.VEC3, ComponentType.FLOAT);
  static AttributeDescription TANGENT = new AttributeDescription(
      MeshAttribute.TANGENT, AccessorType.VEC4, ComponentType.FLOAT);

  AttributeDescription(this.name, this.accessor, this.component);

  /// Returns the [size] in bytes
  int get size {
    int count = 0;
    switch (accessor) {
      case AccessorType.VEC2:
        count = 2;
        break;
      case AccessorType.VEC3:
        count = 3;
        break;
      case AccessorType.VEC4:
        count = 4;
        break;
    }

    int size = 0;
    switch (component) {
      case ComponentType.FLOAT:
        size = 4;
        break;
      case ComponentType.UNSIGNED_BYTE:
        size = 1;
        break;
      case ComponentType.UNSIGNED_SHORT:
        size = 1;
        break;
    }

    return count * size;
  }
}

class Mesh extends NativeFieldWrapperClass2 {
  int _id;
  String name;

  static Mesh _plane;
  static Mesh _cube;
  static Mesh _sphere;

  /// Creates a mesh
  static Future<Mesh> create(
      int vertexCount, int indexCount, List<AttributeDescription> attributes,
      {ComponentType indexComponentType = ComponentType.UNSIGNED_INT,
      bool interleaved = false,
      ByteBuffer vertexData,
      Uint32List indices}) {
    var attributeDesc = new List<int>((attributes.length * 3) + 4);
    attributeDesc[0] = vertexCount;
    attributeDesc[1] = indexCount;
    attributeDesc[2] = interleaved ? 1 : 0;
    attributeDesc[3] = indexComponentType.index;
    int i = 4;
    for (var a in attributes) {
      attributeDesc[i++] = a.name.index;
      attributeDesc[i++] = a.accessor.index;
      attributeDesc[i++] = a.component.index;
    }
    return _MasonService._dispatch(_CREATE_MESH, attributeDesc).then((mesh_id) {
      return new Mesh._withId(mesh_id);
    });
  }

  Mesh._withId(int id) {
    _constructor(id);
  }

  Mesh._primitive(type) {
    _constructor(type);
  }

  _constructor(int id) native "Mesh_constructor";

  Future setVertices(Uint8List vertices) {
    var completer = new Completer();
    SetVerticesNative(vertices, completer);
    return completer.future;
  }

  SetVerticesNative(Uint8List vertices, Completer completer)
      native "Mesh_SetVertices";

  Future setIndices(Uint8List indices) {
    var completer = new Completer();
    SetIndicesNative(indices, completer);
    return completer.future;
  }

  SetIndicesNative(Uint8List indices, Completer completer)
      native "Mesh_SetIndices";

  static Future<Mesh> sphere({double radius = 0.5}) async {
    int X_SEGMENTS = 64;
    int Y_SEGMENTS = 64;

    var positions = [];
    var normals = [];
    var uv = [];

    int vertexSize = 3 + 3 + 2;
    var vertexCount = (X_SEGMENTS + 1) * (Y_SEGMENTS + 1);
    var vertexData = new Float32List(vertexSize * vertexCount);

    for (int y = 0; y <= Y_SEGMENTS; ++y) {
      for (int x = 0; x <= X_SEGMENTS; ++x) {
        double xSegment = x / X_SEGMENTS;
        double ySegment = y / Y_SEGMENTS;
        double xPos =
            Math.cos(xSegment * 2.0 * Math.PI) * Math.sin(ySegment * Math.PI);
        double yPos = Math.cos(ySegment * Math.PI);
        double zPos =
            Math.sin(xSegment * 2.0 * Math.PI) * Math.sin(ySegment * Math.PI);

        positions.add(xPos * radius);
        positions.add(yPos * radius);
        positions.add(zPos * radius);

        uv.add(xSegment);
        uv.add(ySegment);

        var normal = new Vector3(xPos, yPos, zPos)..normalize();
        normals.add(normal.x);
        normals.add(normal.y);
        normals.add(normal.z);
      }
    }

    vertexData.setAll(0, positions);
    vertexData.setAll(positions.length, normals);
    vertexData.setAll(positions.length + normals.length, uv);

    var indices = [];
    for (int y = 0; y < Y_SEGMENTS; ++y) {
      int base1 = (X_SEGMENTS + 1) * y;
      int base2 = (X_SEGMENTS + 1) * (y + 1);

      for (int x = 0; x < X_SEGMENTS; ++x) {
        indices.add(base1 + x);
        indices.add(base1 + x + 1);
        indices.add(base2 + x);

        indices.add(base1 + x + 1);
        indices.add(base2 + x + 1);
        indices.add(base2 + x);
      }
    }

    var indexData = new Uint32List.fromList(indices);

    var mesh = await Mesh.create(vertexCount, indexData.length, [
      AttributeDescription.POSITION,
      AttributeDescription.NORMAL,
      new AttributeDescription(
          MeshAttribute.TEXCOORD_0, AccessorType.VEC2, ComponentType.FLOAT),
    ]);
    mesh.setVertices(vertexData.buffer.asUint8List());
    mesh.setIndices(indexData.buffer.asUint8List());

    return mesh;
  }

  static Future<Mesh> plane(
      {double width = 1.0,
      double height = 1.0,
      Vector3 normal,
      int xDivisions = 1,
      int yDivisions = 1}) async {
    if (normal == null) {
      normal = new Vector3(0.0, 1.0, 0.0);
    }
    if (xDivisions < 1 || yDivisions < 1) {
      throw new Exception("Divisions must be greater than 0");
    }

    normal.normalize();

    int vertexCount = (xDivisions + 1) * (yDivisions + 1);
    int indexCount = (xDivisions * yDivisions) * 2 * 3;

    var mesh = await Mesh.create(vertexCount, indexCount, [
      AttributeDescription.POSITION,
      AttributeDescription.NORMAL,
      new AttributeDescription(
          MeshAttribute.TEXCOORD_0, AccessorType.VEC2, ComponentType.FLOAT),
    ]);

    var vertexData = new Float32List(vertexCount * (3 + 3 + 2));
    var indexData = new Uint32List(indexCount);

    var wHalf = width / 2;
    var hHalf = height / 2;

    var xInc = width / xDivisions;
    var yInc = height / yDivisions;

    var up = Vector3.Y_POSITIVE;
    var m = new Matrix4.identity();
    var angle = normal.dot(up);

    var q = new Quaternion.fromTwoVectors(Vector3.Y_POSITIVE, normal);

    var diff = Vector3.Y_POSITIVE - normal;

    // We start at -x, +z
    var xVertices = xDivisions + 1;
    var yVertices = yDivisions + 1;

    var vertex = 0;

    var normalOffset = 3 * vertexCount;
    var texOffset = normalOffset + 3 * vertexCount;

    for (int i = 0; i < yVertices; i++) {
      for (int j = 0; j < xVertices; j++) {
        double x, y, z;
        x = (j * xInc) - wHalf;
        y = 0.0; // Always zero
        z = height - (i * yInc) - hHalf;

        var position = new Vector3(x, y, z);
        position.applyQuaternion(q);
        vertexData[(vertex * 3) + 0] = position.x;
        vertexData[(vertex * 3) + 1] = position.y;
        vertexData[(vertex * 3) + 2] = position.z;

        //NORMAL
        vertexData[normalOffset + (vertex * 3) + 0] = normal.x;
        vertexData[normalOffset + (vertex * 3) + 1] = normal.y;
        vertexData[normalOffset + (vertex * 3) + 2] = normal.z;

        double u = (j * xInc) / width;
        double v = (i * yInc) / height;

        //TEXCOORD_0
        vertexData[texOffset + (vertex * 2) + 0] = u;
        vertexData[texOffset + (vertex * 2) + 1] = v;

        vertex++;
      }
    }

    for (int i = 0; i < yDivisions; i++) {
      for (int j = 0; j < xDivisions; j++) {
        // j is x
        // i is z
        var vertex = i * xVertices + j;
        var offset = vertex * 6;

        int up = vertex + xVertices;
        int up_over = up + 1;
        int right = vertex + 1;

        indexData[offset] = vertex;
        indexData[offset + 1] = up_over;
        indexData[offset + 2] = up;

        indexData[offset + 3] = vertex;
        indexData[offset + 4] = right;
        indexData[offset + 5] = up_over;
      }
    }

    mesh.setVertices(vertexData.buffer.asUint8List());
    mesh.setIndices(indexData.buffer.asUint8List());

    return mesh;
  }

  static Future<Mesh> cube() async {
    var vertexCount = 24;
    var mesh = await Mesh.create(vertexCount, 36, [
      AttributeDescription.POSITION,
      AttributeDescription.NORMAL,
      new AttributeDescription(
          MeshAttribute.TEXCOORD_0, AccessorType.VEC2, ComponentType.FLOAT),
    ]);

    var width = 1.0;
    var height = 1.0;
    var depth = 1.0;

    var _w = width / 2;
    var _h = height / 2;
    var _d = depth / 2;

    var vertexSize = (3 + 3 + 2);
    var vertexData = new Float32List(vertexCount * vertexSize);
    var indexData = new Uint32List.fromList([
      0,
      1,
      2,
      0,
      2,
      3,
      4,
      5,
      6,
      4,
      6,
      7,
      8,
      9,
      10,
      8,
      10,
      11,
      12,
      13,
      14,
      12,
      14,
      15,
      16,
      17,
      18,
      16,
      18,
      19,
      20,
      21,
      22,
      20,
      22,
      23
    ]);

    int i = 0;

    //Front
    vertexData[i++] = _w;
    vertexData[i++] = _h;
    vertexData[i++] = _d;

    vertexData[i++] = -_w;
    vertexData[i++] = _h;
    vertexData[i++] = _d;

    vertexData[i++] = -_w;
    vertexData[i++] = -_h;
    vertexData[i++] = _d;

    vertexData[i++] = _w;
    vertexData[i++] = -_h;
    vertexData[i++] = _d;

    //Back
    vertexData[i++] = _w;
    vertexData[i++] = -_h;
    vertexData[i++] = -_d;

    vertexData[i++] = -_w;
    vertexData[i++] = -_h;
    vertexData[i++] = -_d;

    vertexData[i++] = -_w;
    vertexData[i++] = _h;
    vertexData[i++] = -_d;

    vertexData[i++] = _w;
    vertexData[i++] = _h;
    vertexData[i++] = -_d;

    //Right
    vertexData[i++] = _w;
    vertexData[i++] = -_h;
    vertexData[i++] = _d;

    vertexData[i++] = _w;
    vertexData[i++] = -_h;
    vertexData[i++] = -_d;

    vertexData[i++] = _w;
    vertexData[i++] = _h;
    vertexData[i++] = -_d;

    vertexData[i++] = _w;
    vertexData[i++] = _h;
    vertexData[i++] = _d;

    //Left
    vertexData[i++] = -_w;
    vertexData[i++] = _h;
    vertexData[i++] = _d;

    vertexData[i++] = -_w;
    vertexData[i++] = _h;
    vertexData[i++] = -_d;

    vertexData[i++] = -_w;
    vertexData[i++] = -_h;
    vertexData[i++] = -_d;

    vertexData[i++] = -_w;
    vertexData[i++] = -_h;
    vertexData[i++] = _d;

    //Top
    vertexData[i++] = _w;
    vertexData[i++] = _h;
    vertexData[i++] = _d;

    vertexData[i++] = _w;
    vertexData[i++] = _h;
    vertexData[i++] = -_d;

    vertexData[i++] = -_w;
    vertexData[i++] = _h;
    vertexData[i++] = -_d;

    vertexData[i++] = -_w;
    vertexData[i++] = _h;
    vertexData[i++] = _d;

    //Bottom
    vertexData[i++] = -_w;
    vertexData[i++] = -_h;
    vertexData[i++] = _d;

    vertexData[i++] = -_w;
    vertexData[i++] = -_h;
    vertexData[i++] = -_d;

    vertexData[i++] = _w;
    vertexData[i++] = -_h;
    vertexData[i++] = -_d;

    vertexData[i++] = _w;
    vertexData[i++] = -_h;
    vertexData[i++] = _d;

    //Normals
    //Front
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;

    //Back
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = -11.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = -1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = -1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = -1.0;

    //Right
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;

    //Left
    vertexData[i++] = -1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = -1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = -1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = -1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;

    //Top
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;

    //Bottom
    vertexData[i++] = 0.0;
    vertexData[i++] = -1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = -1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = -1.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;
    vertexData[i++] = -1.0;
    vertexData[i++] = 0.0;

    //UVs
    //Front
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;

    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;

    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;

    vertexData[i++] = 1.0;
    vertexData[i++] = 1.0;

    //Back
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;

    vertexData[i++] = 1.0;
    vertexData[i++] = 1.0;

    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;

    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;

    //Right
    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;

    vertexData[i++] = 1.0;
    vertexData[i++] = 1.0;

    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;

    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;

    //Left
    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;

    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;

    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;

    vertexData[i++] = 1.0;
    vertexData[i++] = 1.0;

    //Top
    vertexData[i++] = 1.0;
    vertexData[i++] = 1.0;

    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;

    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;

    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;

    //Bottom
    vertexData[i++] = 0.0;
    vertexData[i++] = 0.0;

    vertexData[i++] = 0.0;
    vertexData[i++] = 1.0;

    vertexData[i++] = 1.0;
    vertexData[i++] = 1.0;

    vertexData[i++] = 1.0;
    vertexData[i++] = 0.0;

    mesh.setVertices(vertexData.buffer.asUint8List());
    mesh.setIndices(indexData.buffer.asUint8List());
    return mesh;
  }

  String toString() {
    return "Mesh ID: ${_id}";
  }
}
