part of dart_mason;

class Physics extends NativeFieldWrapperClass2 {
  Vector3 _gravity;

  /// Returns a copy of the current gravity
  Vector get gravity => _gravity.clone();

  Set<CollisionObject> _objects = new Set<CollisionObject>();

  /// Set's the direction an acceleration of gravity.
  /// The vector used is a copy. You can not use a reference to the original
  /// vector to change gravity. You must set the gravity variable again.
  set gravity(Vector3 gravity) {
    _gravity = gravity.clone();
    _setGravityNative(gravity.x, gravity.y, gravity.z);
  }

  Physics({Vector3 gravity}) {
    _constructor();
    gravity ??= new Vector3(0.0, -10.0, 0.0);
    this.gravity = gravity;
  }

  _constructor() native "Physics_constructor";

  /// Steps the physics simulation
  step(Duration time) {
    _stepNative(time.inMicroseconds / Duration.MICROSECONDS_PER_SECOND);
  }

  _stepNative(double seconds) native "Physics_Step";

  // Set's the gravity in the internal bullet simulator.
  _setGravityNative(double x, double y, double z) native "Physics_SetGravity";

  /// Adds a [RigidBody] to the simulation.
  addRigidBody(RigidBody object) {
    if(object._physics != null){
      throw "[RigidBody] is already added to existing [Physics] object.";
    } else {
      _objects.add(object);
//      object._physics = this;
      object._nativeSetPhysicsPtr(this);
      _nativeAddRigidBody(object);
    }
  }

  _nativeAddRigidBody(RigidBody object) native "Physics_AddRigidBody";

  /// Removes a [RigidBody] from the simulation.
  removeRigidBody(RigidBody object) {
    //object._removePhysicsHandler(this);
  }

  /// Adds a [CollisionObject] to the simulation.
  addCollisionObject(CollisionObject object) {
    if(object._physics != null){
      throw "[CollisionObject] is already added to existing [Physics] object.";
    } else {
      _objects.add(object);
//      object._physics = this;
      object._nativeSetPhysicsPtr(this);
      _nativeAddCollisionObject(object);
    }
  }

  _nativeAddCollisionObject(CollisionObject object)
      native "Physics_AddCollisionObject";

  /// Remove [CollisionObject] from the simulation
  removeCollisionObject(CollisionObject object) {}

  /// Casts a ray in the given direction [Vector3]
  List<_RaycastHit> rayTest(Vector3 origin, Vector3 direction,
      [double distance = 10.0]) {
    Vector3 dir = direction.clone()..normalize();
    Vector3 to = origin + (dir * distance);
    return _nativeRayTest(origin, to);
  }

  _RaycastHit rayTestClosest(Vector3 origin, Vector3 direction,
      [double distance = 10.0]) {
    Vector3 dir = direction.clone()..normalize();
    Vector3 to = origin + (dir * distance);
    return _nativeRayTestClosest(origin, to);
  }

  List<_RaycastHit> _nativeRayTest(Vector3 from, Vector3 to)
      native "Physics_RayTest";
  _RaycastHit _nativeRayTestClosest(Vector3 from, Vector3 to)
      native "Physics_RayTestClosest";
}

/// Creates a physics object that can detect collisions.
class CollisionObject extends NativeFieldWrapperClass2 {
  Physics _physics;
  _CollisionShape shape;

  Transform _transform = new Transform();

  CollisionObject(this.shape, {Transform transform}) {
    _constructor(_CollisionObjectType.COLLISION_OBJECT.index);
    _nativeInit();
    _nativeSetCollisionShape(this.shape);
    if (transform != null) {
      setTransform(transform);
    }
  }

  CollisionObject._(_CollisionObjectType type) {
    _constructor(type.index);
  }

  _constructor(int type) native "CollisionObject_constructor";
  _nativeInit() native "CollisionObject_Init";
  _nativeSetCollisionShape(_CollisionShape)
      native "CollisionObject_SetCollisionShape";

  /// Set's the transform for the [CollisionObject]
  /// Scale is ignored for physics simulations. Scale the [_CollisionShape]
  /// instead of the [CollisionObject]
  setTransform(Transform transform) {
    _transform = transform;
    _nativeSetTransform(transform.position, transform.rotation);
  }

  _nativeSetTransform(Vector3 position, Quaternion rotation)
      native "CollisionObject_SetTransform";

  _nativeSetPhysicsPtr(Physics physics) native "CollisionObject_SetPhysicsPtr";

  /// Get's the current transform of the [CollisionObject]
  Transform get transform {
    return _transform;
  }
}

typedef void OnUpdateMotionState();
typedef void OnSetMotionState();

class MotionState {
  MotionState({OnUpdateMotionState onUpdate, OnSetMotionState onSet});
}

typedef void RigidBodyOnStep(Transform transform);

class RigidBody extends CollisionObject {
  _CollisionShape _shape;
  RigidBodyOnStep onStep;

  StreamController<Transform> _physicsUpdate =
      new StreamController<Transform>();

  ///
  Stream<Transform> get onUpdate => _physicsUpdate.stream.asBroadcastStream();

  RigidBody(double mass, this._shape, {this.onStep})
      : super._(_CollisionObjectType.RIGID_BODY) {
    _nativeRigidBodyConstructor(mass, _shape);
  }

  _nativeRigidBodyConstructor(double mass, _CollisionShape shape)
      native "RigidBody_Constructor";

  updatePhysics(Vector3 position, Quaternion rotation) {
    _physicsUpdate.add(new Transform(position: position, rotation: rotation));
  }

  void onStepInner(Transform transform){
    _transform = transform;
    if(onStep != null){
      onStep(transform);
    }
  }
}

// Enums mush match collision_shapes.cc
enum _CollisionShapeType {
  BOX,
  SPHERE,
  CONE,
  CAPSULE,
  CYLINDER,
  PLANE,
}
enum _CollisionObjectType {
  COLLISION_OBJECT,
  RIGID_BODY,
}

abstract class _CollisionShape extends NativeFieldWrapperClass2 {
  _CollisionShape(_CollisionShapeType type) {
    _constructor(type.index);
  }

  _constructor(int type) native "CollisionShape_constructor";

  init() native "CollisionShape_Init";

  setMargin(double margin);

  setScale(Vector3 scale) native "CollisionShape_SetScale";
}

class BoxCollider extends _CollisionShape {
  BoxCollider(Vector3 size) : super(_CollisionShapeType.BOX) {
    _boxConstructor(size);
  }

  _boxConstructor(Vector3 size) native "BoxShape_BoxConstructor";

  setMargin(double margin) native "BoxShape_SetMargin";
}

class SphereCollider extends _CollisionShape {
  SphereCollider(double radius) : super(_CollisionShapeType.SPHERE) {
    _sphereConstructor(radius);
  }

  _sphereConstructor(double radius) native "SphereShape_SphereConstructor";

  setMargin(double margin) native "SphereShape_SetMargin";
}

class ConeCollider extends _CollisionShape {
  ConeCollider(double radius, double height) : super(_CollisionShapeType.CONE) {
    _coneConstructor(radius, height);
  }

  _coneConstructor(double radius, double height)
      native "ConeShape_ConeConstructor";

  setMargin(double margin) native "ConeShape_SetMargin";
}

class CapsuleCollider extends _CollisionShape {
  CapsuleCollider(double radius, double height)
      : super(_CollisionShapeType.CAPSULE) {
    _capsuleConstructor(radius, height);
  }

  _capsuleConstructor(double radius, double height)
      native "CapsuleShape_CapsuleConstructor";

  setMargin(double margin) native "CapsuleShape_SetMargin";
}

class CylinderCollider extends _CollisionShape {
  CylinderCollider(double radius, double height)
      : super(_CollisionShapeType.CYLINDER) {
    _cylinderConstructor(radius, height);
  }

  _cylinderConstructor(double radius, double height)
      native "CylinderShape_CylinderConstructor";

  setMargin(double margin) native "CapsuleShape_SetMargin";
}

class _PlaneCollider extends _CollisionShape {
  _PlaneCollider._(Axis axis) : super(_CollisionShapeType.PLANE) {
    _planeConstructor(axis.index);
  }

  _planeConstructor(int axis) native "PlaneShape_PlaneConstructor";

  setMargin(double margin) native "PlaneShape_SetMargin";
}

/// Creates a 1x1 Plane.
class PlaneXCollider extends _PlaneCollider {
  PlaneXCollider() : super._(Axis.X_POSITIVE);
}
class PlaneYCollider extends _PlaneCollider {
  PlaneYCollider() : super._(Axis.Y_POSITIVE);
}
class PlaneZCollider extends _PlaneCollider {
  PlaneZCollider() : super._(Axis.Z_POSITIVE);
}

class _RaycastHit {
  CollisionObject collider;
  Vector3 normal;
  Vector3 point;

  _RaycastHit(this.point, this.normal, this.collider);
}
