library dart_materials;
// ignore_for_file: native_function_body_in_non_sdk_code,extends_non_class,undefined_class


import 'dart:_internal';
import 'dart:async';
import 'dart:convert';
import 'dart:developer' as developer;
import 'dart:math' as Math;
import 'dart:nativewrappers';
import 'dart:typed_data';
import 'dart:isolate';
import 'dart:mirrors';
import 'dart:mason';

part 'material_service.dart';

class _ParamDetails<T>{
  final String uniformKey;
  final T defaultValue;

  _ParamDetails(this.uniformKey, this.defaultValue);
}

/// Handle to a compiled shader.
/// To use call createInstance()
class Material extends NativeFieldWrapperClass2 {
  /// Handle to shader id on glass side.

  final int _id;

  Map<String, _ParamDetails<Texture>> _textures = new Map<String, _ParamDetails<Texture>>();
  Map<String, _ParamDetails<Object>> _params = new Map<String, _ParamDetails<Object>>();

  Material._(this._id, this._params, this._textures){
    _constructor(_id);
  }


  _constructor(int id) native "Material_constructor";

  MaterialInstance createInstance([Map<String, Object> params = const {}]){
    var m = new _MaterialInstance(this);
      params.forEach((key, value){
        m.setParam(key, value);
      });
    return m;
  }

  _addTexture(String textureName, Texture texture){
    _textures[textureName] = texture;
  }

  _addParam(String paramName, Object defaultValue){
    _params[paramName] = defaultValue;
  }

  Map<String, Object> get textures => _textures;
  Map<String, Object> get params => _params;

}

class MaterialInstance{
  Material get material;
  Map<String, Object> get params;
  setParam(String key, Object value);
}

class _MaterialInstance extends MaterialInstance{
  final Material _material;

  Map<String, Object> _params = new Map<String, Object>();

  _MaterialInstance(this._material);

  setParam(String key, Object value){
    _params[key] = value;
  }

  Map<String, Object> get params => _params;
  Material get material => _material;
}

class _MaterialNode{

}

class MaterialBuilder {
  Object _baseColor;
  Object _metallic;
  Object _roughness;
  Object _ao;
  Object _normal;

  int _nextParamNumber = 1;

  set baseColor(Object color) {
    if (color is Vector3) {
      _baseColor = color;
    } else if (color is Vector4) {
      _baseColor = color;
    } else if (color is Texture) {
      _baseColor = color;
    } else if (color is _MaterialNode) {
      _baseColor = color;
    } else {
      throw new Error("Invalid type.");
    }
  }
  set metallic(Object metallic) {
    if (metallic is double) {
      _metallic = metallic;
    } else if(metallic is _MaterialNode){
      _metallic = metallic;
    } else {
      throw new Error("Invalid type.");
    }
  }
  
  set roughness(Object roughness) {
    if (roughness is double) {
      _roughness = roughness;
    } else if(roughness is _MaterialNode){
      _roughness = roughness;
    } else {
      throw new Error("Invalid type.");
    }
  }

  set ao(Object ao) {
    if (ao is double) {
      _ao = ao;
    } else if(ao is _MaterialNode){
      _ao = ao;
    } else {
      throw new Error("Invalid type.");
    }
  }

//  set normal(Object normal) {
//    if(normal is _MaterialNode){
//      _normal = normal;
//    } else {
//      throw new Error("Invalid type.");
//    }
//  }

  Future<Material> compile() {
    // Need to produce the string that will go into the uniform section
    // and that will go into the compute section.
    String uniforms = "";
    String compute = "";
    int fixed = 5;

    var textures = new Map<String, _ParamDetails>();
    var params = new Map<String, _ParamDetails>();

    /// Base Color
    if(_baseColor is Vector3){
      compute += "vec3 albedo = vec3(${_baseColor.x.toStringAsFixed(fixed)}, ${_baseColor.y.toStringAsFixed(fixed)}, ${_baseColor.z.toStringAsFixed(fixed)}, 0.0);\n";
      compute += "float alpha = 1.0;\n";
    } else if (_baseColor is Vector4){
      compute += "vec3 albedo = vec3(${_baseColor.x.toStringAsFixed(fixed)}, ${_baseColor.y.toStringAsFixed(fixed)}, ${_baseColor.z.toStringAsFixed(fixed)});\n";
      compute += "float alpha = ${_baseColor.w.toStringAsFixed(fixed)};\n";
    } else if(_baseColor is _ParamNode) {
      if(_baseColor is TextureSamplerParam){
        var textureName = _getParamName();
        uniforms += "uniform sampler2D ${textureName};\n"; //TODO Generate a texture name.
        params[_baseColor._key] = new _ParamDetails(textureName, _baseColor.texture);
        compute += "vec3 albedo = pow(texture(${textureName}, TexCoords).rgb, vec3(5.2));\n";
        compute += "float alpha = texture(${textureName}, TexCoords).a;\n";
      }
      if(_baseColor is Vector4Param){
        var paramName = _getParamName();
        uniforms += "uniform vec4 ${paramName};\n"; //TODO Generate a texture name.
        params[_baseColor._key] = new _ParamDetails(paramName, _baseColor._defaultValue);
        compute += "vec3 albedo = ${paramName}.xyz;\n";
        compute += "float alpha = ${paramName}.a;\n";
      }
    } else if (_baseColor is _MaterialNode) {
      if (_baseColor is TextureSampler) {
        var textureName = _getParamName();
        uniforms += "uniform sampler2D ${textureName};\n"; //TODO Generate a texture name.
        textures[textureName] =
        new _ParamDetails(textureName, _baseColor.texture);
        compute +=
        "vec3 albedo = pow(texture(${textureName}, TexCoords).rgb, vec3(5.2));\n";
        compute += "float alpha = texture(${textureName}, TexCoords).a;\n";
      }
    } else if(_baseColor == null){
      compute += "vec3 albedo = vec3(0.0, 0.0, 1.0);";
    }else {
      throw new Error("Must be a Vector, num, or _MaterialNode");
    }

    /// Metallic
    if(_metallic is double) {
      compute += "float metallic = ${_metallic.toStringAsFixed(fixed)};\n";
    } else if (_metallic is _ParamNode){
      if(_metallic is ScalarParam) {
        var paramName = _getParamName();
        uniforms += "uniform float ${paramName};\n";
        params[_metallic._key] =
        new _ParamDetails(paramName, _metallic._getDefaultValue());
        compute += "float metallic = ${paramName};\n";
      } else if(_metallic is TextureSamplerParam){
        var textureName = _getParamName();
        uniforms += "uniform sampler2D ${textureName};\n"; //TODO Generate a texture name.
        params[_metallic._key] = new _ParamDetails(textureName, _metallic.texture);
        compute += "float metallic = texture(${textureName}, TexCoords).r;\n";
      }
    } else if (_metallic is _MaterialNode){
      if(_metallic is TextureSampler){
        var textureName = _getParamName();
        uniforms += "uniform sampler2D ${textureName};\n";//TODO Generate a texture name.
        textures[textureName] = new _ParamDetails(textureName, _metallic.texture);
        compute += "float metallic = texture(${textureName}, TexCoords).r;\n";
      }
    } else if (_metallic == null){
      compute += "float metallic = 1.0;\n";
    } else {
      throw new Error("Must be a Vector, num, or _MaterialNode");
    }
    
    /// Roughness
    if(_roughness is double) {
      compute += "float roughness = ${_roughness.toStringAsFixed(fixed)};\n";
    } else if (_roughness is _ParamNode){
      if(_roughness is ScalarParam){
        var paramName = _getParamName();
        uniforms += "uniform float ${paramName};\n";
        params[_roughness._key] = new _ParamDetails(paramName, _roughness._getDefaultValue());
        compute += "float roughness = ${paramName};\n";
      } else if(_roughness is TextureSamplerParam){
        var textureName = _getParamName();
        uniforms += "uniform sampler2D ${textureName};\n"; //TODO Generate a texture name.
        params[_roughness._key] = new _ParamDetails(textureName, _roughness.texture);
        compute += "float roughness = texture(${textureName}, TexCoords).r;\n";
      }
    } else if (_roughness is _MaterialNode) {
      if (_roughness is TextureSampler) {
        var textureName = _getParamName();
        uniforms +=
        "uniform sampler2D ${textureName};\n"; //TODO Generate a texture name.
        textures[textureName] =
        new _ParamDetails(textureName, _roughness.texture);
        compute += "float roughness = texture(${textureName}, TexCoords).r;\n";
      }
    } else if (_roughness == null){
      compute += "float roughness = 1.0;\n";
    } else {
      throw new Error("Must be a Vector, num, or _MaterialNode");
    }
    
    /// AO
    if(_ao is double) {
      compute += "float ao = ${_ao.toStringAsFixed(fixed)};\n";
    } else if (_ao is _ParamNode){
      if(_ao is ScalarParam){
        var paramName = _getParamName();
        uniforms += "uniform float ${paramName};\n";
        params[_ao._key] = new _ParamDetails(paramName, _ao._getDefaultValue());
        compute += "float ao = ${paramName};\n";
      } else if(_ao is TextureSamplerParam){
        var textureName = _getParamName();
        uniforms += "uniform sampler2D ${textureName};\n"; //TODO Generate a texture name.
        params[_ao._key] = new _ParamDetails(textureName, _ao.texture);
        compute += "float ao = texture(${textureName}, TexCoords).r;\n";
      }
    } else if (_ao is _MaterialNode) {
      if (_ao is TextureSampler) {
        var textureName = _getParamName();
        uniforms +=
        "uniform sampler2D ${textureName};\n"; //TODO Generate a texture name.
        textures[textureName] = new _ParamDetails(textureName, _ao.texture);
        compute += "float ao = texture(${textureName}, TexCoords).r;\n";
      }
    } else if(_ao == null){
      compute += "float ao = 1.0;\n";
    } else {
      throw new Error("Must be a Vector, num, or _MaterialNode");
    }

    //Normal
    if(_normal is _MaterialNode){
      if(_normal is TextureSampler){
        var textureName = _getParamName();
        uniforms += "uniform sampler2D ${textureName};\n";//TODO Generate a texture name.
        textures[textureName] = new _ParamDetails(textureName, _ao.texture);
//        compute += "vec3 N = getNormalFromMap(${textureName});";
        compute += "vec3 N = texture(${textureName}, vec2(TexCoords.x, -TexCoords.y)).rgb;";
        compute += "N = normalize(N * 2.0 - 1.0);";
      }
    } else if(_normal == null) {
      compute += "vec3 N = normalize(ModelNormal);";
    } else {
      throw new Error("Must be a Vector, num, or _MaterialNode");
    }

    return _MaterialService._dispatch(_MATERIAL_COMPILE, [uniforms, compute]).then((int shader_id) {
      var material = new Material._(shader_id, params, textures);
      return material;
    });
    //return new Future.value(material);
  }

  String _getParamName(){
    return "param${_nextParamNumber++}";
  }
}

class TextureSampler extends _MaterialNode{
  final Texture texture;
  TextureSampler(this.texture){

  }
}

class TextureSamplerParam extends _ParamNode{
  final Texture texture;
  TextureSamplerParam(String key, this.texture):super(key){

  }

  Texture _getDefaultValue(){
    return texture;
  }
}

class _ParamNode extends _MaterialNode{
  String _key;
  _ParamNode(this._key);

  Object _getDefaultValue();
}

class ScalarParam extends _ParamNode{
  final double _defaultValue;

  ScalarParam(String key, this._defaultValue): super(key);

  double _getDefaultValue(){
    return _defaultValue;
  }
}


class Vector2Param extends _ParamNode{
  Vector2 _defaultValue;
  Vector2Param(String key, this._defaultValue): super(key);
  Vector2 _getDefaultValue(){
    return _defaultValue;
  }
}
class Vector3Param extends _ParamNode{
  final Vector3 _defaultValue;
  Vector3Param(String key, this._defaultValue): super(key);
  Vector3 _getDefaultValue(){
    return _defaultValue;
  }
}
class Vector4Param extends _ParamNode{
  final Vector4 _defaultValue;
  Vector4Param(String key, this._defaultValue): super(key);
  Vector4 _getDefaultValue(){
    return _defaultValue;
  }
}
class TextureParam extends _ParamNode{
  final Texture _defaultTexture;
  TextureParam(String key, this._defaultTexture): super(key);
  Texture _getDefaultValue(){
    return _defaultTexture;
  }
}

