part of solarium.io;

abstract class HttpMethod {
  static const String GET = "get";
  static const String DELETE = "delete";
  static const String POST = "post";
  static const String PUT = "put";
  static const String PATCH = "patch";
  static const String HEAD = "head";
}

abstract class HttpClient {
  static const int DEFAULT_HTTP_PORT = 80;
  static const int DEFAULT_HTTPS_PORT = 443;

  factory HttpClient() => new _HttpClient();

  Future<HttpResponse> head(url, {HttpHeaders headers});

  Future<HttpResponse> get(url, {HttpHeaders headers});

  Future<HttpResponse> post(url, body, {HttpHeaders headers});

  Future<HttpResponse> put(url, body, {HttpHeaders headers});

  Future<HttpResponse> patch(url, body, {HttpHeaders headers});

  Future<HttpResponse> delete(url, {HttpHeaders headers});

  Future<HttpResponse> send(HttpRequest request);
}

abstract class HttpRequest {
  Uri get url;

  String get method;

  HttpHeaders headers;

  factory HttpRequest(method, url) => new _HttpRequest(method, url);

  Future<HttpResponse> send();
}

/// The body of a response
class Body {
  Encoding _encoding;

  Stream<List<int>> _stream;

  Body._(this._stream, [this._encoding]);

  Future<String> read([Encoding encoding]) {
    encoding ??= this._encoding ?? UTF8;
    return encoding.decodeStream(readBytes());
  }

  Stream<List<int>> readBytes() {
    if (_stream == null) {
      throw new StateError("The 'read' method can only be called once on a "
          "http.Request/http.Response object.");
    }
    var stream = _stream;
    _stream = null;
    return stream;
  }
}

abstract class HttpResponse {
  /// The location of the requested resource.
  ///
  /// The value takes into account any redirects that occurred during the
  /// request.
  Uri get finalUrl;

  int get status;
  String get reasonPhrase;
  Body get body;
  HttpHeaders get headers;
}

class _HttpClient implements HttpClient {
  SendPort port;

  @override
  Future<HttpResponse> delete(url, {HttpHeaders headers}) {
    // TODO: implement delete
  }

  @override
  Future<HttpResponse> get(url, {HttpHeaders headers}) {
    var request = new _HttpRequestHandler();
    return request.Open(getUrl(url)).then((reader) {
      var response = request.ReadResponseHeaders();
      return response;
    });
  }

  @override
  Future<HttpResponse> head(url, {HttpHeaders headers}) {
    // TODO: implement head
  }

  @override
  Future<HttpResponse> patch(url, body, {HttpHeaders headers}) {
    // TODO: implement patch
  }

  @override
  Future<HttpResponse> post(url, body, {HttpHeaders headers}) {
    // TODO: implement post
  }

  @override
  Future<HttpResponse> put(url, body, {HttpHeaders headers}) {
    // TODO: implement put
  }

  @override
  Future<HttpResponse> send(HttpRequest request) {
    // TODO: implement send
  }
}

class _HttpRequestHandler {
  int pointer;

  StreamController<List<int>> streamController =
      new StreamController<List<int>>();

  Future<_HttpResponseReader> Open(String url) {
    //Method and headers also?
    return _IOService._dispatch(_HTTP_OPEN, [url]).then((response) {
      pointer = response;
      return pointer;
    });
  }

  Future<bool> WriteBytes(List<int> bytes) {
    return _IOService._dispatch(_HTTP_WRITE_BYTES, [pointer]).then((response) {
      return true;
    });
  }

  Future<HttpResponse> ReadResponseHeaders() {
    return _IOService._dispatch(_HTTP_READ_RESPONSE_HEADER, [pointer]).then(
        (List io_response) {
      Uint8List headers_data = io_response[0];
      int net_error = io_response[1];
      String finalUrl = io_response[2];
      if(net_error == 0){ // Success
        print(net_error);
        //Parse the headers.
        var header_lines = [];
        var temp = [];
        for (var i = 0; i < headers_data.length; i++) {
          if (headers_data[i] == 0) {
            var line = ASCII.decode(temp);
            header_lines.add(line);
            temp = [];
          } else {
            temp.add(headers_data[i]);
          }
        }
        //Last line should always be empty, remove.
        header_lines.removeLast();

        var status = _parseStatus(header_lines[0]);
        header_lines = header_lines.sublist(1);
        var headers = _HttpHeaders.parse(header_lines);
        var body = new Body._(streamController.stream);
        var reader = new _HttpResponseReader(pointer, streamController);
        var response = new _HttpResponse(body, status[0], status[1], headers, Uri.parse(finalUrl), reader);
        return response;
      } else {
        throw new Exception("Network Error.");
      }
    });
  }

  bool isDone = false;
  void StartReadBody() {
    while (!isDone) {
      ReadBody();
    }
  }

  void ReadBody() {
    _IOService._dispatch(_HTTP_READ_BODY, [pointer]).then((List<int> response) {
      if (response.length == 0) {
        Close();
        streamController.close();
        isDone = true;
      } else {
        streamController.add(response);
      }
    });
  }

  Future Close() {
    return _IOService._dispatch(_HTTP_CLOSE, [pointer]).then((response) {
      //No need to return anything here.
    });
  }

  // Parses a line such as "HTTP/1.1 200 OK"
  List _parseStatus(String line) {
    var statusNum;
    var statusMessage;
    var index = line.indexOf("HTTP/1.1");
    if (index == 0) {
      var status = line.substring("HTTP/1.1".length).trim();
      int ptr = 0;
      while (ptr < status.length && int.parse(status[ptr], onError: (source) => null) != null) {
        ++ptr;
      }
      statusNum = int.parse(status.substring(0, ptr));
      statusMessage = status.substring(ptr).trim();
    }
    return [statusNum, statusMessage];
  }
}

class _HttpResponseReader {
  final int pointer;
  final StreamController<List<int>> streamController;

  _HttpResponseReader(this.pointer, this.streamController);

  Uint8List ReadData() => ReadDataNative(pointer);
  Uint8List ReadDataNative(int pointer) native "HttpResponseReader_Read";
  Close(int pointer) native "HttpResponseReader_Close";

  void ReadDataUntilComplete() {
    var keepReading = true;
    while (keepReading) {
      var data = ReadData();
      if (data.length == 0) {
        keepReading = false;
        streamController.close();
        Close(pointer);
      } else {
        streamController.add(data);
      }
    }
  }
}

class _HttpRequest implements HttpRequest {
  final String method;
  final Uri url;

  @override
  HttpHeaders headers;

  _HttpRequest(this.method, this.url) {}

  @override
  send() {}
}

class _HttpResponse implements HttpResponse {
  // TODO: implement body
  final Body body;

  final int status;

  final String reasonPhrase;

  final HttpHeaders headers;

  final Uri finalUrl;

  final _HttpResponseReader _reader;

  _HttpResponse(this.body, this.status, this.reasonPhrase, this.headers, this.finalUrl, this._reader) {
    this._reader.ReadDataUntilComplete();
  }

}

/**
 * HTTP status codes.
 */
abstract class HttpStatus {
  static const int CONTINUE = 100;
  static const int SWITCHING_PROTOCOLS = 101;
  static const int OK = 200;
  static const int CREATED = 201;
  static const int ACCEPTED = 202;
  static const int NON_AUTHORITATIVE_INFORMATION = 203;
  static const int NO_CONTENT = 204;
  static const int RESET_CONTENT = 205;
  static const int PARTIAL_CONTENT = 206;
  static const int MULTIPLE_CHOICES = 300;
  static const int MOVED_PERMANENTLY = 301;
  static const int FOUND = 302;
  static const int MOVED_TEMPORARILY = 302; // Common alias for FOUND.
  static const int SEE_OTHER = 303;
  static const int NOT_MODIFIED = 304;
  static const int USE_PROXY = 305;
  static const int TEMPORARY_REDIRECT = 307;
  static const int BAD_REQUEST = 400;
  static const int UNAUTHORIZED = 401;
  static const int PAYMENT_REQUIRED = 402;
  static const int FORBIDDEN = 403;
  static const int NOT_FOUND = 404;
  static const int METHOD_NOT_ALLOWED = 405;
  static const int NOT_ACCEPTABLE = 406;
  static const int PROXY_AUTHENTICATION_REQUIRED = 407;
  static const int REQUEST_TIMEOUT = 408;
  static const int CONFLICT = 409;
  static const int GONE = 410;
  static const int LENGTH_REQUIRED = 411;
  static const int PRECONDITION_FAILED = 412;
  static const int REQUEST_ENTITY_TOO_LARGE = 413;
  static const int REQUEST_URI_TOO_LONG = 414;
  static const int UNSUPPORTED_MEDIA_TYPE = 415;
  static const int REQUESTED_RANGE_NOT_SATISFIABLE = 416;
  static const int EXPECTATION_FAILED = 417;
  static const int INTERNAL_SERVER_ERROR = 500;
  static const int NOT_IMPLEMENTED = 501;
  static const int BAD_GATEWAY = 502;
  static const int SERVICE_UNAVAILABLE = 503;
  static const int GATEWAY_TIMEOUT = 504;
  static const int HTTP_VERSION_NOT_SUPPORTED = 505;

  // Client generated status code.
  static const int NETWORK_CONNECT_TIMEOUT_ERROR = 599;
}

/**
 * Headers for HTTP requests and responses.
 *
 * In some situations, headers are immutable:
 *
 * * HttpRequest and HttpClientResponse always have immutable headers.
 *
 * * HttpResponse and HttpClientRequest have immutable headers
 *   from the moment the body is written to.
 *
 * In these situations, the mutating methods throw exceptions.
 *
 * For all operations on HTTP headers the header name is
 * case-insensitive.
 *
 * To set the value of a header use the `set()` method:
 *
 *     request.headers.set(HttpHeaders.CACHE_CONTROL,
 *                         'max-age=3600, must-revalidate');
 *
 * To retrieve the value of a header use the `value()` method:
 *
 *     print(request.headers.value(HttpHeaders.USER_AGENT));
 *
 * An HttpHeaders object holds a list of values for each name
 * as the standard allows. In most cases a name holds only a single value,
 * The most common mode of operation is to use `set()` for setting a value,
 * and `value()` for retrieving a value.
 */
abstract class HttpHeaders {
  static const ACCEPT = "accept";
  static const ACCEPT_CHARSET = "accept-charset";
  static const ACCEPT_ENCODING = "accept-encoding";
  static const ACCEPT_LANGUAGE = "accept-language";
  static const ACCEPT_RANGES = "accept-ranges";
  static const AGE = "age";
  static const ALLOW = "allow";
  static const AUTHORIZATION = "authorization";
  static const CACHE_CONTROL = "cache-control";
  static const CONNECTION = "connection";
  static const CONTENT_ENCODING = "content-encoding";
  static const CONTENT_LANGUAGE = "content-language";
  static const CONTENT_LENGTH = "content-length";
  static const CONTENT_LOCATION = "content-location";
  static const CONTENT_MD5 = "content-md5";
  static const CONTENT_RANGE = "content-range";
  static const CONTENT_TYPE = "content-type";
  static const DATE = "date";
  static const ETAG = "etag";
  static const EXPECT = "expect";
  static const EXPIRES = "expires";
  static const FROM = "from";
  static const HOST = "host";
  static const IF_MATCH = "if-match";
  static const IF_MODIFIED_SINCE = "if-modified-since";
  static const IF_NONE_MATCH = "if-none-match";
  static const IF_RANGE = "if-range";
  static const IF_UNMODIFIED_SINCE = "if-unmodified-since";
  static const LAST_MODIFIED = "last-modified";
  static const LOCATION = "location";
  static const MAX_FORWARDS = "max-forwards";
  static const PRAGMA = "pragma";
  static const PROXY_AUTHENTICATE = "proxy-authenticate";
  static const PROXY_AUTHORIZATION = "proxy-authorization";
  static const RANGE = "range";
  static const REFERER = "referer";
  static const RETRY_AFTER = "retry-after";
  static const SERVER = "server";
  static const TE = "te";
  static const TRAILER = "trailer";
  static const TRANSFER_ENCODING = "transfer-encoding";
  static const UPGRADE = "upgrade";
  static const USER_AGENT = "user-agent";
  static const VARY = "vary";
  static const VIA = "via";
  static const WARNING = "warning";
  static const WWW_AUTHENTICATE = "www-authenticate";

  // Cookie headers from RFC 6265.
  static const COOKIE = "cookie";
  static const SET_COOKIE = "set-cookie";

  static const GENERAL_HEADERS = const [
    CACHE_CONTROL,
    CONNECTION,
    DATE,
    PRAGMA,
    TRAILER,
    TRANSFER_ENCODING,
    UPGRADE,
    VIA,
    WARNING
  ];

  static const ENTITY_HEADERS = const [
    ALLOW,
    CONTENT_ENCODING,
    CONTENT_LANGUAGE,
    CONTENT_LENGTH,
    CONTENT_LOCATION,
    CONTENT_MD5,
    CONTENT_RANGE,
    CONTENT_TYPE,
    EXPIRES,
    LAST_MODIFIED
  ];

  static const RESPONSE_HEADERS = const [
    ACCEPT_RANGES,
    AGE,
    ETAG,
    LOCATION,
    PROXY_AUTHENTICATE,
    RETRY_AFTER,
    SERVER,
    VARY,
    WWW_AUTHENTICATE
  ];

  static const REQUEST_HEADERS = const [
    ACCEPT,
    ACCEPT_CHARSET,
    ACCEPT_ENCODING,
    ACCEPT_LANGUAGE,
    AUTHORIZATION,
    EXPECT,
    FROM,
    HOST,
    IF_MATCH,
    IF_MODIFIED_SINCE,
    IF_NONE_MATCH,
    IF_RANGE,
    IF_UNMODIFIED_SINCE,
    MAX_FORWARDS,
    PROXY_AUTHORIZATION,
    RANGE,
    REFERER,
    TE,
    USER_AGENT
  ];

  /**
     * Gets and sets the date. The value of this property will
     * reflect the 'date' header.
     */
  DateTime date;

  /**
     * Gets and sets the expiry date. The value of this property will
     * reflect the 'expires' header.
     */
  DateTime expires;

  /**
     * Gets and sets the "if-modified-since" date. The value of this property will
     * reflect the "if-modified-since" header.
     */
  DateTime ifModifiedSince;

  /**
     * Gets and sets the host part of the 'host' header for the
     * connection.
     */
  String host;

  /**
     * Gets and sets the port part of the 'host' header for the
     * connection.
     */
  int port;

  /**
     * Gets and sets the content type. Note that the content type in the
     * header will only be updated if this field is set
     * directly. Mutating the returned current value will have no
     * effect.
     */
  ContentType contentType;

  /**
     * Gets and sets the content length header value.
     */
  int contentLength;

  factory HttpHeaders() => new _HttpHeaders();

  /**
     * Returns the list of values for the header named [name]. If there
     * is no header with the provided name, [:null:] will be returned.
     */
  List<String> operator [](String name);

  /**
     * Convenience method for the value for a single valued header. If
     * there is no header with the provided name, [:null:] will be
     * returned. If the header has more than one value an exception is
     * thrown.
     */
  String value(String name);

  /**
     * Adds a header value. The header named [name] will have the value
     * [value] added to its list of values. Some headers are single
     * valued, and for these adding a value will replace the previous
     * value. If the value is of type DateTime a HTTP date format will be
     * applied. If the value is a [:List:] each element of the list will
     * be added separately. For all other types the default [:toString:]
     * method will be used.
     */
  void add(String name, Object value);

  /**
     * Sets a header. The header named [name] will have all its values
     * cleared before the value [value] is added as its value.
     */
  void set(String name, Object value);

  /**
     * Removes a specific value for a header name. Some headers have
     * system supplied values and for these the system supplied values
     * will still be added to the collection of values for the header.
     */
  void remove(String name, Object value);

  /**
     * Removes all values for the specified header name. Some headers
     * have system supplied values and for these the system supplied
     * values will still be added to the collection of values for the
     * header.
     */
  void removeAll(String name);

  /**
     * Enumerates the headers, applying the function [f] to each
     * header. The header name passed in [:name:] will be all lower
     * case.
     */
  void forEach(void f(String name, List<String> values));

  /**
     * Remove all headers. Some headers have system supplied values and
     * for these the system supplied values will still be added to the
     * collection of values for the header.
     */
  void clear();
}

/**
 * Representation of a header value in the form:
 *
 *   [:value; parameter1=value1; parameter2=value2:]
 *
 * [HeaderValue] can be used to conveniently build and parse header
 * values on this form.
 *
 * To build an [:accepts:] header with the value
 *
 *     text/plain; q=0.3, text/html
 *
 * use code like this:
 *
 *     HttpClientRequest request = ...;
 *     var v = new HeaderValue("text/plain", {"q": "0.3"});
 *     request.headers.add(HttpHeaders.ACCEPT, v);
 *     request.headers.add(HttpHeaders.ACCEPT, "text/html");
 *
 * To parse the header values use the [:parse:] static method.
 *
 *     HttpRequest request = ...;
 *     List<String> values = request.headers[HttpHeaders.ACCEPT];
 *     values.forEach((value) {
 *       HeaderValue v = HeaderValue.parse(value);
 *       // Use v.value and v.parameters
 *     });
 *
 * An instance of [HeaderValue] is immutable.
 */
abstract class HeaderValue {
  /**
     * Creates a new header value object setting the value and parameters.
     */
  factory HeaderValue([String value = "", Map<String, String> parameters]) {
    return new _HeaderValue(value, parameters);
  }

  /**
     * Creates a new header value object from parsing a header value
     * string with both value and optional parameters.
     */
  static HeaderValue parse(String value,
      {String parameterSeparator: ";",
      String valueSeparator: null,
      bool preserveBackslash: false}) {
    return _HeaderValue.parse(value,
        parameterSeparator: parameterSeparator,
        valueSeparator: valueSeparator,
        preserveBackslash: preserveBackslash);
  }

  /**
     * Gets the header value.
     */
  String get value;

  /**
     * Gets the map of parameters.
     *
     * This map cannot be modified. invoking any operation which would
     * modify the map will throw [UnsupportedError].
     */
  Map<String, String> get parameters;

  /**
     * Returns the formatted string representation in the form:
     *
     *     value; parameter1=value1; parameter2=value2
     */
  String toString();
}

/**
 * Representation of a content type. An instance of [ContentType] is
 * immutable.
 */
abstract class ContentType implements HeaderValue {
  /**
     * Content type for plain text using UTF-8 encoding.
     *
     *     text/plain; charset=utf-8
     */
  static final TEXT = new ContentType("text", "plain", charset: "utf-8");

  /**
     *  Content type for HTML using UTF-8 encoding.
     *
     *     text/html; charset=utf-8
     */
  static final HTML = new ContentType("text", "html", charset: "utf-8");

  /**
     *  Content type for JSON using UTF-8 encoding.
     *
     *     application/json; charset=utf-8
     */
  static final JSON = new ContentType("application", "json", charset: "utf-8");

  /**
     *  Content type for binary data.
     *
     *     application/octet-stream
     */
  static final BINARY = new ContentType("application", "octet-stream");

  /**
     * Creates a new content type object setting the primary type and
     * sub type. The charset and additional parameters can also be set
     * using [charset] and [parameters]. If charset is passed and
     * [parameters] contains charset as well the passed [charset] will
     * override the value in parameters. Keys passed in parameters will be
     * converted to lower case. The `charset` entry, whether passed as `charset`
     * or in `parameters`, will have its value converted to lower-case.
     */
  factory ContentType(String primaryType, String subType,
      {String charset, Map<String, String> parameters}) {
    return new _ContentType(primaryType, subType, charset, parameters);
  }

  /**
     * Creates a new content type object from parsing a Content-Type
     * header value. As primary type, sub type and parameter names and
     * values are not case sensitive all these values will be converted
     * to lower case. Parsing this string
     *
     *     text/html; charset=utf-8
     *
     * will create a content type object with primary type [:text:], sub
     * type [:html:] and parameter [:charset:] with value [:utf-8:].
     */
  static ContentType parse(String value) {
    return _ContentType.parse(value);
  }

  /**
     * Gets the mime-type, without any parameters.
     */
  String get mimeType;

  /**
     * Gets the primary type.
     */
  String get primaryType;

  /**
     * Gets the sub type.
     */
  String get subType;

  /**
     * Gets the character set.
     */
  String get charset;
}

/**
 * Representation of a cookie. For cookies received by the server as
 * Cookie header values only [:name:] and [:value:] fields will be
 * set. When building a cookie for the 'set-cookie' header in the server
 * and when receiving cookies in the client as 'set-cookie' headers all
 * fields can be used.
 */
abstract class Cookie {
  /**
     * Gets and sets the name.
     */
  String name;

  /**
     * Gets and sets the value.
     */
  String value;

  /**
     * Gets and sets the expiry date.
     */
  DateTime expires;

  /**
     * Gets and sets the max age. A value of [:0:] means delete cookie
     * now.
     */
  int maxAge;

  /**
     * Gets and sets the domain.
     */
  String domain;

  /**
     * Gets and sets the path.
     */
  String path;

  /**
     * Gets and sets whether this cookie is secure.
     */
  bool secure;

  /**
     * Gets and sets whether this cookie is HTTP only.
     */
  bool httpOnly;

  /**
     * Creates a new cookie optionally setting the name and value.
     *
     * By default the value of `httpOnly` will be set to `true`.
     */
  factory Cookie([String name, String value]) => new _Cookie(name, value);

  /**
     * Creates a new cookie by parsing a header value from a 'set-cookie'
     * header.
     */
  factory Cookie.fromSetCookieValue(String value) {
    return new _Cookie.fromSetCookieValue(value);
  }

  /**
     * Returns the formatted string representation of the cookie. The
     * string representation can be used for for setting the Cookie or
     * 'set-cookie' headers
     */
  String toString();
}

class HttpException implements IOException {
  final String message;
  final Uri uri;

  const HttpException(this.message, {this.uri});

  String toString() {
    var b = new StringBuffer()..write('HttpException: ')..write(message);
    if (uri != null) {
      b.write(', uri = $uri');
    }
    return b.toString();
  }
}

/// Utils

/// Returns a [Uri] from the [url], which can be a [Uri] or a [String].
///
/// If the [url] is not a [Uri] or [String] an [ArgumentError] is thrown.
String getUrl(url) {
  if (url is Uri) {
    return url.toString();
  } else if (url is String) {
    return url;
  } else {
    throw new ArgumentError.value(url, 'url', 'Not a Uri or String');
  }
}
